#include "PainterEngine_Application.h"
#include "pearl_white_story.h"

volatile px_bool lock=PX_FALSE;
PX_Application App;

px_void PX_MidiCallbackTrigger(struct _PX_Midi* midi, px_byte opcode, px_byte  track, px_byte  instrument, px_byte Note, px_byte velocity, px_void* userptr)
{
	px_char content[8];
	PX_Application* pApp = (PX_Application*)userptr;
	PX_Object_Piano* pPiano = PX_ObjectGetDesc(PX_Object_Piano, pApp->piano_keyboard);
	if (opcode==0x90)
	{
		px_int index;
		PX_MidiNoteToString(Note, content);
		index = PX_PianoKeyNameToIndex(content);
		PX_PianoTriggerIndex(&pApp->Piano, index, velocity / 127.0f);
		pPiano->keyState[index] = PX_OBJECT_PIANO_KEYSTATE_DOWN;
	}

	if (opcode == 0x80)
	{
		px_int index;
		PX_MidiNoteToString(Note, content);
		index = PX_PianoKeyNameToIndex(content);
		PX_PianoTriggerIndex(&pApp->Piano, index, 0);
		pPiano->keyState[index] = PX_OBJECT_PIANO_KEYSTATE_STANDBY;
	}
}

px_void PX_ApplicationOnPianoKey(PX_Object* pObject, PX_Object_Event e, px_void* ptr)
{
	PX_Application* pApp = (PX_Application*)ptr;
	px_int index= PX_Object_Event_GetIndex(e);
	pApp->currentSelectKey = index;
	PX_PianoTriggerIndex(&pApp->Piano, index, 1);
	PX_Object_PianoTuneSetParameters(pApp->piano_tune, &pApp->Piano,index);
	PX_Object_PianoClearKeyFocus(pApp->piano_keyboard);
	PX_Object_PianoSetKeyFocus(pApp->piano_keyboard, index);
}

px_void PX_ApplicationOnPianoPlay(PX_Object* pObject, PX_Object_Event e, px_void* ptr)
{
	PX_Application* pApp = (PX_Application*)ptr;
	PX_MidiReset(&pApp->midi);
	PX_MidiPlay(&pApp->midi);
}

px_void PX_ApplicationOnPianoStop(PX_Object* pObject, PX_Object_Event e, px_void* ptr)
{
	PX_Application* pApp = (PX_Application*)ptr;
	PX_MidiPause(&pApp->midi);
}

px_void PX_ApplicationOnPianoReset(PX_Object* pObject, PX_Object_Event e, px_void* ptr)
{
	PX_Application* pApp = (PX_Application*)ptr;
	if (pApp->currentSelectKey!=-1)
	{
		PX_PianoKey_Parameters param;
		const px_float PX_Piano_KeyFrequency[] = { 27.5f,29.135f,30.868f,32.703f,34.648f,36.708f,38.891f,41.203f,43.654f,46.249f,48.999f,51.913f };
		param.f = PX_Piano_KeyFrequency[pApp->currentSelectKey % 12] * PX_pow_ii(2, pApp->currentSelectKey / 12);
		param.Fs = 44100;
		param.weight = 1;
		param.minr = 0.35f;
		param.maxr = 2;
		param.amprl = 4;
		param.amprr = 8;
		param.mult_radius_core_string = 1;
		param.minL = 0.07f;
		param.maxL = 1.4f;
		param.ampLl = -4.f;
		param.ampLr = 4.f;
		param.mult_density_string = 1;
		param.mult_modulus_string = 1;
		param.mult_impedance_bridge = 1;
		param.mult_impedance_hammer = 0;
		param.mult_mass_hammer = 1;
		param.mult_force_hammer = 0.2f;
		param.mult_hysteresis_hammer = 1;
		param.mult_stiffness_exponent_hammer = 1;
		param.position_hammer = 0.142f;
		param.mult_loss_filter = 1;
		param.detune = 0.0003f;
		param.hammer_type = 1;
		PX_PianoKeyFree(&pApp->Piano.keys[pApp->currentSelectKey]);
		PX_PianoKeyInitialize(&pApp->runtime.mp_game,&pApp->Piano.keys[pApp->currentSelectKey],param);
		PX_Object_PianoTuneSetParameters(pApp->piano_tune, &pApp->Piano, pApp->currentSelectKey);
	}
	
}

px_void PX_ApplicationOnPianoSave(PX_Object* pObject, PX_Object_Event e, px_void* ptr)
{
	px_int i;
	PX_Application* pApp = (PX_Application*)ptr;
	px_memory configData;
	PX_MemoryInitialize(&pApp->runtime.mp_game, &configData);
	for (i=0;i<88;i++)
	{
		PX_MemoryCat(&configData, &pApp->Piano.keys[i].param, sizeof(PX_PianoKey_Parameters));
	}
	PX_MemoryCat(&configData, &pApp->Piano.soundboard.param, sizeof(PX_PianoSoundboard_Parameters));
	PX_SaveDataToFile(configData.buffer, configData.usedsize, "config.piano");
	PX_Object_MessageBoxAlertOk(pApp->messagebox, "Save piano data to current path \"config.piano\"", 0, 0);
	

}

px_void PX_ApplicationOnPianoLoad(PX_Object* pObject, PX_Object_Event e, px_void* ptr)
{
	PX_IO_Data io;
	PX_Application* pApp = (PX_Application*)ptr;
	io = PX_LoadFileToIOData("config.piano");
	if (io.size!= sizeof(PX_PianoKey_Parameters) * 88 + sizeof(PX_PianoSoundboard_Parameters))
	{
		PX_Object_MessageBoxAlertOk(pApp->messagebox, "Could not found(or invalid) \"config.piano\"", 0, 0);
		return;
	}
	if (!lock)
	{
		PX_PianoKey_Parameters* pkeyparam;
		PX_PianoSoundboard_Parameters * psoundboardparam;
		lock = PX_TRUE;
		pkeyparam = (PX_PianoKey_Parameters*)io.buffer;
		psoundboardparam = (PX_PianoSoundboard_Parameters*)(io.buffer + sizeof(PX_PianoKey_Parameters) * 88);
		PX_PianoFree(&pApp->Piano);
		PX_PianoInitializeEx(&pApp->runtime.mp_game, &pApp->Piano, pkeyparam, *psoundboardparam);
		lock = PX_FALSE;
		PX_Object_MessageBoxAlertOk(pApp->messagebox, "Load piano configurations succeeded", 0, 0); 
		if (pApp->currentSelectKey != -1)
		{
			PX_Object_PianoTuneSetParameters(pApp->piano_tune, &pApp->Piano, pApp->currentSelectKey);
		}
	}
	
	PX_FreeIOData(&io);
	

}
px_void PX_PianoRead(px_void* userptr, px_byte* pBuffer, px_int readSize)
{
	if (!readSize||readSize>2048*2*2)
	{
		PX_memset(pBuffer, 0, readSize);
	}
	else
	{
		px_int oft = 0;
		PX_Application* pApp = (PX_Application*)userptr;
		px_int count = readSize / 4;//2bytes 2channels
		px_float out44100f[2048];
		px_int i;
		px_short* PCM = (px_short*)pBuffer;
		if (!lock)
		{
			lock = PX_TRUE;
			PX_PianoGo(&pApp->Piano, out44100f, count);
			lock = PX_FALSE;
		}
			
		for (i=0;i<count;i++)
		{
			PCM[oft++] = (px_short)(out44100f[i]*32768*8);
			PCM[oft++] = (px_short)(out44100f[i] * 32768*8);
		}
	}
}

px_void PX_ApplicationOnPianoTuneChanged(PX_Object* pObject, PX_Object_Event e, px_void* ptr)
{
	PX_Application* pApp = (PX_Application*)ptr;
	PX_Object_PianoTune* pTune = PX_ObjectGetDesc(PX_Object_PianoTune, pApp->piano_tune);
	PX_PianoKey* pKey;
	if (pApp->currentSelectKey==-1)
	{
		return;
	}
	pKey = &pApp->Piano.keys[pApp->currentSelectKey];
	if (!PX_memequ(&pTune->key_param,&pKey->param,sizeof(PX_PianoKey_Parameters)))
	{
		if (!lock)
		{
			lock = PX_TRUE;
			PX_PianoKeyFree(pKey);
			if (!PX_PianoKeyInitialize(&pApp->runtime.mp_game, pKey, pTune->key_param))
				PX_ERROR("out of memory");
			lock = PX_FALSE;
		}
	}

	if (!PX_memequ(&pTune->soundboard_param, &pApp->Piano.soundboard, sizeof(PX_PianoSoundboard_Parameters)))
	{
		if (!lock)
		{
			lock = PX_TRUE;
			PX_PianoSoundBoardFree(&pApp->Piano.soundboard);
			if (!PX_PianoSoundBoardInitialize(&pApp->runtime.mp_game, &pApp->Piano.soundboard, pTune->soundboard_param))
				PX_ERROR("out of memory");
			lock = PX_FALSE;
		}
	}
}


px_bool PX_ApplicationInitialize(PX_Application *pApp,px_int screen_width,px_int screen_height)
{

	PX_ApplicationInitializeDefault(&pApp->runtime, screen_width, screen_height);
	PX_MidiInitialize(&pApp->midi, &pApp->runtime.mp_resources, PX_MidiCallbackTrigger, pApp);
	PX_MidiLoad(&pApp->midi, pearl_white_story, sizeof(pearl_white_story));
	PX_MidiSetQuarterNoteDuration(&pApp->midi, 640);
	PX_PianoInitialize( &pApp->runtime.mp_game, &pApp->Piano,PX_PIANO_STYLE_DEBUG);
	PX_SoundPlayInitialize(&pApp->runtime.mp_game, &pApp->runtime.soundplay);
	PX_AudioInitialize(&pApp->runtime.soundplay);
	PX_SoundPlaySetUserRead(&pApp->runtime.soundplay, PX_PianoRead, pApp);
	pApp->root = PX_ObjectCreate(&pApp->runtime.mp_ui, 0, 0, 0, 0, 0, 0, 0);
	pApp->piano_keyboard = PX_Object_PianoCreate(&pApp->runtime.mp_ui, pApp->root, 0, 0, 52*16, 64);
	pApp->button_play = PX_Object_PushButtonCreate(&pApp->runtime.mp_ui, pApp->root, 265, 5, 96, 32, "Play Test", PX_NULL);
	pApp->button_stop= PX_Object_PushButtonCreate(&pApp->runtime.mp_ui, pApp->root, 365, 5, 96, 32, "Stop Test", PX_NULL);
	pApp->button_reset = PX_Object_PushButtonCreate(&pApp->runtime.mp_ui, pApp->root, 465, 5, 96, 32, "Reset Key", PX_NULL);
	pApp->button_save = PX_Object_PushButtonCreate(&pApp->runtime.mp_ui, pApp->root, 565, 5, 96, 32, "Save Config", PX_NULL);
	pApp->button_load = PX_Object_PushButtonCreate(&pApp->runtime.mp_ui, pApp->root, 665, 5, 96, 32, "Load Config", PX_NULL);
	pApp->piano_tune = PX_Object_PianoTuneCreate(&pApp->runtime.mp_ui, pApp->root, 0, 0, 100, 100);
	pApp->messagebox = PX_Object_MessageBoxCreate(&pApp->runtime.mp_ui, pApp->root, PX_NULL);
	pApp->currentSelectKey = -1;
	PX_ObjectRegisterEvent(pApp->piano_keyboard, PX_OBJECT_EVENT_VALUECHANGED, PX_ApplicationOnPianoKey, pApp);
	PX_ObjectRegisterEvent(pApp->button_play, PX_OBJECT_EVENT_EXECUTE, PX_ApplicationOnPianoPlay, pApp);
	PX_ObjectRegisterEvent(pApp->button_stop, PX_OBJECT_EVENT_EXECUTE, PX_ApplicationOnPianoStop, pApp);
	PX_ObjectRegisterEvent(pApp->button_reset, PX_OBJECT_EVENT_EXECUTE, PX_ApplicationOnPianoReset, pApp);
	PX_ObjectRegisterEvent(pApp->button_save, PX_OBJECT_EVENT_EXECUTE, PX_ApplicationOnPianoSave, pApp);
	PX_ObjectRegisterEvent(pApp->button_load, PX_OBJECT_EVENT_EXECUTE, PX_ApplicationOnPianoLoad, pApp);
	PX_ObjectRegisterEvent(pApp->piano_tune, PX_OBJECT_EVENT_VALUECHANGED, PX_ApplicationOnPianoTuneChanged, pApp);
	
	return PX_TRUE;
}

px_void PX_ApplicationUpdate(PX_Application *pApp,px_dword elapsed)
{
	PX_MidiUpdate(&pApp->midi, elapsed);
	
}

px_void PX_ApplicationRender(PX_Application *pApp,px_dword elapsed)
{
	px_surface *pRenderSurface=&pApp->runtime.RenderSurface;
	PX_RuntimeRenderClear(&pApp->runtime,PX_OBJECT_UI_DEFAULT_BACKGROUNDCOLOR);
	pApp->piano_keyboard->Width = pRenderSurface->width * 19 / 20.f;
	pApp->piano_keyboard->Height = 64;
	pApp->piano_keyboard->x = pRenderSurface->width / 2 - pApp->piano_keyboard->Width / 2;
	pApp->piano_keyboard->y = pRenderSurface->height - pApp->piano_keyboard->Height ;

	pApp->piano_tune->x = 0;
	pApp->piano_tune->y = 24;
	pApp->piano_tune->Width = pRenderSurface->width*1.0f;
	pApp->piano_tune->Height = pRenderSurface->height - 128.f;

	PX_ObjectRender(pRenderSurface, pApp->root, elapsed);

	PX_FontDrawText(pRenderSurface, 5, 3, PX_ALIGN_LEFTTOP, "PainterEngine", PX_OBJECT_UI_DEFAULT_FONTCOLOR);
	PX_FontDrawText(pRenderSurface, 12, 20, PX_ALIGN_LEFTTOP, "Physics-based piano Tuner", PX_COLOR_WHITE);
	
}

px_void PX_ApplicationPostEvent(PX_Application *pApp,PX_Object_Event e)
{
	PX_ApplicationEventDefault(&pApp->runtime, e);
	PX_ObjectPostEvent( pApp->root, e);
}

